# checker - plainchk - slurmchk - sslurmchk - dslurmchk - rgdchk

import os

from ..plugins import DSlurmChk, DSlurmSub
from ..config import conf, history


class RgDSub(DSlurmSub):
    pass


class RgDChk(DSlurmChk):
    def _render_input(self, jobid, checkid="", param=None, prefix=""):
        if not prefix:
            super()._render_input(jobid, checkid, param, prefix)
        else:  # check input
            with open(os.path.join(conf["inputs_abs_dir"], jobid), "r") as f:
                _, l = f.readlines()
            l = float(l)
            with open(os.path.join(conf["check_inputs_abs_dir"], checkid), "w") as f:
                f.writelines(["%s\n%s" % (l, param)])

    def check_checking_main(self, jobid):
        with open(
            os.path.join(
                conf["work_dir"], conf["check_outputs_dir"], history[jobid]["assoc"]
            ),
            "r",
        ) as f:
            b = f.readlines()
        b = float(b[0])
        if b < 0:
            return []
        else:
            with open(os.path.join(conf["inputs_abs_dir"], jobid), "r") as f:
                L, l = f.readlines()
            L = float(L)
            l = float(l)
            return [(L, l)]

    def _render_check_input(self, jobid, checkid, param):
        with open(os.path.join(conf["inputs_abs_dir"], jobid), "r") as f:
            _, l = f.readlines()
        l = float(l)
        with open(
            os.path.join(conf["work_dir"], conf["check_inputs_dir"], checkid), "w"
        ) as f:
            f.writelines(["%s\n%s" % (l, param)])
        # check script should be line2 - line1

    def check_finished_main(self, jobid):
        with open(os.path.join(conf["outputs_abs_dir"], jobid), "r") as f:
            r = f.readlines()
        r = float(r[0])

        return [r]  # check_param



"""
Basic classes: submitter, checker and processor for subway main loop
"""

import os
import sys
import heapq
from abc import ABC, abstractmethod

from .utils import now_ts, md5file
from .config import history, conf, update_history, update_conf
from .exceptions import EndingBubble


class Checker(ABC):
    """
    Empty base abstract class for checker.
    """

    @abstractmethod
    def __call__(self):
        pass


class Submitter(ABC):
    """
    Empty base abstract class for submitter.
    """

    @abstractmethod
    def __call__(self):
        pass


class Processor(ABC):
    """
    Base class for processor
    """

    def __init__(self, pipeline=None, **kwargs):
        """

        :param pipeline: List[str]. List of str of method name that processor will run one by one.
        :param kwargs:
        """
        if not pipeline:
            pipeline = []
        self.pipeline = pipeline
        self.kws = kwargs

    def __call__(self):
        for func in self.pipeline:
            getattr(self, func)()


class PlainChk(Checker):
    """
    The very general checker class.
    """

    def __init__(self, params=None, **kwargs):
        """

        :param params: list of dict
        """
        self.kws = kwargs
        self.params = params

    def __call__(self):
        """
        `DIY: not recommended`.
        main process of checker.

        :return: None.
        """
        if not history:  ## preprocess for kickstart inputs file and history.json

            nfs = self.check_task()
            for nf, resource in nfs:
                self.post_new_input(nf, resource)

        else:
            running_fs = [f for f, s in history.items() if s["state"] == "running"]
            if not running_fs:
                print("no job is running", file=sys.stderr)
            for f in running_fs:
                if self.is_finished(f) or self.is_aborted(f):
                    history[f]["finishing_ts"] = self.finishing_time(f)
                    history[f]["state"] = (
                        "finished" if self.is_finished(f) else "aborted"
                    )
                    nfs = self.check_task(f)
                    if nfs:
                        assert (
                            len(nfs) == 1
                        )  # single job for check, one-one correspondence
                        history[f]["assoc"] = nfs[0][0]
                        history[f]["check_resource"] = nfs[0][1]
                    # for nf, resource in nfs:
                    #     self.post_new_input(nf, resource, f)

            checking_fs = [f for f, s in history.items() if s["state"] == "checking"]
            if not checking_fs:
                print("no job is checking", file=sys.stderr)
            for f in checking_fs:
                if self.is_checked(f):
                    nfs = self.check_task(f)
                    if history[f]["state"] != "frustrated":
                        # check_checking also has the right to label the job frustrated
                        history[f]["state"] = "checked"
                    history[f]["ending_ts"] = self.ending_time(f)
                    for nf, resource in nfs:
                        self.post_new_input(nf, resource, f)
                elif self.is_frustrated(f):
                    history[f]["state"] = "frustrated"
                    history[f]["ending_ts"] = self.ending_time(f)
            # else:
            #     print("no checking job is finishing", file=sys.stderr)

            resolving_fs = [f for f, s in history.items() if s["state"] == "resolving"]
            if resolving_fs:
                print("some jobs are resolving", file=sys.stderr)
            for f in resolving_fs:
                if self.is_resolved(f):
                    nfs = self.check_task(f)
                    if history[f]["state"] != "failed":
                        history[f]["state"] = "resolved"
                    history[f]["ending_ts"] = self.ending_time(f)
                    for nf, resource in nfs:
                        self.post_new_input(nf, resource, f)
                elif self.is_failed(f):
                    history[f]["state"] = "failed"
                    history[f]["ending_ts"] = self.ending_time(f)
            # else:
            #     print("no resolving job is finishing", file=sys.stderr)

        update_history()  # update history.json
        ## check whether all tasks are finished, if yes exit main workflow
        end_states = ["failed", "resolved", "checked", "frustrated"]
        for j, s in history.items():
            if s["state"] not in end_states:
                break
        else:
            if not self.kws.get("test"):  # debug for pytest doesn't exit
                raise EndingBubble(message="all jobs are done")

    def post_new_input(self, jobid, resource=None, prev=None):
        """
        `DIY: not recommended.`
        Make some inplace changes on history dict after new task is generated,
        it is vital for history record system.
        This method is not recommended to customize in subclass of checker,
        at least, one should include it by ``super().post_new_input``.

        :param jobid: str. new job id.
        :param resource: Optional[Dict]. resource dict for the new job.
        :param prev: Optional[str]. Parent job id. Default None indicates no prev job.
        :return: None.
        """
        if prev:
            history[prev]["next"].append(jobid)
        history[jobid] = {}
        history[jobid]["state"] = "pending"
        history[jobid]["prev"] = prev
        history[jobid]["next"] = []
        history[jobid]["creating_ts"] = now_ts()
        if not resource:
            resource = {}
        history[jobid]["resource"] = resource
        # resource is in general a dict indicating specific computation resource for this task

    def check_task(self, jobid=None):
        """
        `DIY: not recommended.`
        Determine whether output is "converged" and if not, generate new input files for more calculation
        with the new jobid and resource returned.
        The inner process will dispatch to :py:meth:`check_finished`, :py:meth:`check_aborted`
        :py:meth:`check_checking`, :py:meth:`check_resolving`.

        :param jobid: Optional[str, None]. Default None means kickstart generation for inputs tasks.
        :return: List[Tuple[str, Dict]]. [(next jobid, resource dict)]
        """
        # jobid = None for kickstart,
        # jobid state can be:
        # finished (to generate check input and check id -> checked),
        # aborted (to generate analyse input and check id -> analysed)
        # checked (to generate main input and new id, -> ended)
        # analysed (to generate main input and new id, -> resolved/failed)

        r = None  # [()]
        if not jobid:
            r = self.check_kickstart()
            return r

        s = history[jobid]["state"]

        # if s == "finished":
        #     r = self.check_finished(jobid)
        # elif s == "aborted":
        #     r = self.check_aborted(jobid)
        # elif s == "checking":
        #     r = self.check_checking(jobid)
        # elif s == "resolving":
        #     r = self.check_resolving(jobid)

        if s in ["finished", "aborted", "checking", "resolving"]:
            r = getattr(self, "check_" + s)(jobid)
        else:
            raise ValueError("check_task doesn't support job state %s" % s)
            # TODO: change to subway exceptions?
        return r

    def check_finished(self, jobid):
        """
        `DIY: depends.`
        Check job whose state is ``finished``.
        Namely, the main task has been computed sucessfully,
        we need to check output of main task to further
        generate check job and associated check input.
        On the other hand, if there is no need for check job
        to be controlled by submitter (i.e. one simply check
        the main output in checker and generate new jobs),
        one can omit this method, since the default implementation
        is used for such case (no check job scenario).

        :param jobid: str.
        :return: List[Tuple[str, Dict]]. The length of the list must be
            0 (no check job) or
            1 (namely check job must be one to one correspondence with main job).
        """
        history[jobid]["state"] = "checking"
        history[jobid]["checking_ts"] = now_ts()
        # skip check job submit, check them directly
        return []

    def check_aborted(self, jobid):
        """
        `DIY: depends.`
        Check job whose state is ``aborted``.
        These jobs failed the main task, and this method
        should generate resolve task and resolve input if necessary.
        If resolve job is not under control of submitter,
        the default implementation is ok.

        :param jobid: str.
        :return: List[Tuple[str, Dict]]. The length of the list must be 0 or 1.
        """
        history[jobid]["state"] = "resolving"
        history[jobid]["checking_ts"] = now_ts()
        return []

    def check_kickstart(self):
        """
        `DIY: strongly recommended.`
        Generate jobs with inputs at the beginning.
        The default impl does nothing. If the user doesn't define this method
        in subclass, then the user must add inputs files and possible items
        in empty history.json by hand.

        :return: List[Tuple[str, Dict]]. [(jobid, resource dict)]
        """
        return []

    @abstractmethod
    def check_checking(self, jobid):
        """
        `DIY: must.`
        Check job whose state is ``checking``.
        Jobs are sent to this method only when :py:meth:`is_checked` returns ``True``.
        Generate possible new jobs and inputs based on results from main task
        and possibly check task. The very core of checker class.

        :param jobid: str.
        :return: List[Tuple[str, Dict]]. [(jobid, resource dict), ...]
        """
        return []

    def check_resolving(self, jobid):
        """
        `DIY: depends.`
        Check job whose state is ``resolving``.
        Jobs are sent to this method only when :py:meth:`is_resolved` returns ``True``.
        If the user case has no consideration on error tolerant and aborted states,
        the method can be left as it is doing nothing.

        :param jobid: str.
        :return: List[Tuple[str, Dict]]. [(jobid, resource dict), ...]
        """
        return []

    def is_finished(self, jobid):
        """
        `DIY: strongly recommended.`
        Whether a ``running`` task is finished.
        The default impl is to check whether output file exists.
        It is not a good criteria for finished job in most cases.

        :param jobid: str.
        :return:
        """
        outputs_abs_dir = conf.get("outputs_abs_dir")
        if outputs_abs_dir and os.path.exists(outputs_abs_dir):
            output_fs = [
                f
                for f in os.listdir(outputs_abs_dir)
                if os.path.isfile(os.path.join(conf["outputs_abs_dir"], f))
            ]
            return jobid in output_fs
        return True

    def is_aborted(self, jobid):
        """
        `DIY: depends.`
        Whether a ``running`` task is failed.

        :param jobid: str.
        :return: bool.
        """
        return False

    def is_checked(self, jobid):
        """
        `DIY: depends.`
        Whether a ``checking`` task is finished.

        :param jobid: str.
        :return: bool.
        """
        return True

    def is_frustrated(self, jobid):
        """
        `DIY: depends.`
        Whether a ``checking`` task is failed.

        :param jobid: str.
        :return: bool.
        """
        return False

    def is_resolved(self, jobid):
        """
        `DIY: depends.`
        Whether a ``resolving`` task is finished.

        :param jobid: str.
        :return: bool.
        """
        return True

    def is_failed(self, jobid):
        """
        `DIY: depends.`
        Whether a ``resolving`` task is failed.

        :param jobid: str.
        :return: bool.
        """
        return False

    def finishing_time(self, jobid):
        """
        `DIY: depends.`
        This method is reponsible for 2 states: ``finished`` and ``aborted``.
        The default impl for ``finished`` state is last modifed time of output file.

        :param jobid: str.
        :return: float. timestamp.
        """
        if history[jobid]["state"] == "finished":
            outputs_abs_dir = conf.get("outputs_abs_dir")
            if outputs_abs_dir and os.path.exists(
                os.path.join(conf["outputs_abs_dir"], jobid)
            ):
                return os.path.getmtime(os.path.join(conf["outputs_abs_dir"], jobid))
            return now_ts()
        elif history[jobid]["state"] == "aborted":
            return now_ts()

    def ending_time(self, jobid):
        """
        `DIY: depends.`
        This method is reponsible for 4 states:
        ``checked``, ``resolved``, ``frustrated``, ``failed``.
        The default impl is now.

        :param jobid: str.
        :return: float. timestamp.
        """
        return now_ts()


## creating_ts pending beginning_ts running finishing_ts finished checking_ts checking  ending_ts checked/frustrated
##                                                       aborted checking_ts resolving  ending_ts resolved/failed
class PlainSub(Submitter):
    """
    The very general submitter class.
    """

    def __init__(self, **kwargs):
        self.pending_queue = []
        self.finished_queue = []
        self.aborted_queue = []
        self.is_queue = False
        self.kws = kwargs

    def __call__(self):
        """
        `DIY: not recommend.`

        :return:
        """
        self.is_queue = False  ## fix of realtime update for submitter
        while not self.is_restricted():
            t = self.priority_task()
            # TODO: considering priority and resource together to schedule, not into this recently
            if t is None:
                break
            self.submit_job(t)
            if history[t]["state"] == "pending":
                history[t]["state"] = "running"
                history[t]["beginning_ts"] = now_ts()
                if conf.get("executable_version"):
                    history[t]["executable_version"] = conf["executable_version"]
            elif history[t]["state"] == "finished":
                history[t]["state"] = "checking"
                history[t]["checking_ts"] = now_ts()
                if conf.get("check_executable_version"):
                    history[t]["check_executable_version"] = conf[
                        "check_executable_version"
                    ]
            elif history[t]["state"] == "aborted":
                history[t]["state"] = "resolving"
                history[t]["checking_ts"] = now_ts()
                if conf.get("check_executable_version"):
                    history[t]["check_executable_version"] = conf[
                        "check_executable_version"
                    ]
            else:
                raise ValueError("submit job doesn't support job state %s" % s)

        update_history()

    def priority_task(self):
        """
        `DIY: depends.`
        The default impl prefer jobs in aborted state, then finished state and finally pending state.
        Plus, the default impl prefer older jobs.
        choose the first task to be run from pending, finished and aborted queue

        :return: str, jobid.
        """
        if self.is_queue is False:
            for f, s in history.items():
                if s["state"] == "pending":
                    heapq.heappush(self.pending_queue, (s["creating_ts"], f))
                elif s["state"] == "finished":
                    heapq.heappush(self.finished_queue, (s["finishing_ts"], f))
                elif s["state"] == "aborted":
                    heapq.heappush(self.aborted_queue, (s["finishing_ts"], f))
            self.is_queue = True
        if self.aborted_queue:
            return heapq.heappop(self.aborted_queue)[1]
        elif self.finished_queue:
            return heapq.heappop(self.finished_queue)[1]
        elif self.pending_queue:
            return heapq.heappop(self.pending_queue)[1]
        else:
            return

    def is_restricted(self):
        """
        `DIY: depends.`
        The default impl is ok if you only want to restrict some resource in simple and soft way.
        check the restriction is ok for now with all running tasks in consideration.

        :return: bool. whether the total resource limit is reached.
        """
        # the default pattern: for each job, possible keys are *_count in history resource or check_resource
        # compare these with items *_limit in resource_limit in config
        resource_status = {}
        for j, s in history.items():
            if s["state"] == "running":
                for k, v in s.get("resource", {}).items():
                    if k.endswith("_count"):  # eg. cpu_count: 2
                        resource_status[k] = resource_status.get(k, 0) + v
            elif s["state"] in ["checking", "resolving"]:
                for k, v in s.get("check_resource", {}).items():
                    if k.endswith("_count"):  # eg. cpu_count: 2
                        resource_status[k] = resource_status.get(k, 0) + v
        for res_count, usage in resource_status.items():
            if res_count[:-5] + "limit" in conf.get("resource_limit", {}):
                if conf["resource_limit"][res_count[:-5] + "limit"] <= usage:
                    return True  ## some usage has already bypass the limit
        return False  ## no resource limitation
        # note the implementation here is for soft resource limit, since job ends only when limit is over

    def submit_job(self, jobid):
        """
        `DIY: not recommend`.
        submit job, for example, generate sbatch file and submit it to slurm.
        This method automaticall dispatch to :meth:`submit_pending`, :meth:`submit_finished` and :meth:`submit_aborted`.
        
        :param jobid: str.
        :return: None.
        """
        s = history[jobid]["state"]
        if s == "pending":
            self.submit_pending(jobid)
        elif s == "finished":
            self.submit_finished(jobid)
        elif s == "aborted":
            self.submit_aborted(jobid)
        else:
            raise ValueError("submit job doesn't support job state %s" % s)
        pass

    @abstractmethod
    def submit_pending(self, jobid):
        """
        `DIY: must.`
        The very core of submitter class. It describes how do you define submitting a job.

        :param jobid: str.
        :return:
        """
        pass

    def submit_finished(self, jobid):
        """
        `DIY: depends.`
        If you have to deal with DS scheme as in :ref:`dsss`, then you must implement this method.

        :param jobid: str.
        :return:
        """
        pass

    def submit_aborted(self, jobid):
        """
        `DIY: depends.`
        If you have to deal with DS scheme as in :ref:`dsss` together with fault tolerance, then you must implement this method.

        :param jobid: str.
        :return:
        """
        pass


class PreProcessor(Processor):
    def version_check(self):
        """
        Must also include :meth:`update_conf` into the ``pipeline``.
        Lock the binary version with hashing, ensure different binary cannot be run
        unless the version is changed in ``config.json``/

        :return: None
        """
        if self.kws.get("version_checking_path"):
            cp = self.kws.get("version_checking_path")
            # [(relative path, version number)]
        else:
            cp = []
            for k, v in conf.items():
                if k.endswith("executable") and conf.get(k + "_version"):
                    cp.append((v, conf[k + "_version"]))

        conf.setdefault("hashing", {})
        # hashing: {"main.o": {"1.0.0": "md5agsdhj"}}
        for p in cp:
            conf["hashing"].setdefault(p[0], {})
            f = os.path.join(conf["work_dir"], p[0])
            if os.path.exists(f):
                nhash = md5file(f)
                ohash = conf["hashing"].get(p[0], {}).get(p[1])
                if ohash:
                    assert (
                        nhash == ohash
                    )  # TODO: the exception is to be wrapped in a better presentation
                conf["hashing"][p[0]][p[1]] = nhash

    @staticmethod
    def conf_update():
        """
        Write ``conf`` in memory back to ``config.json``.

        :return: None.
        """
        update_conf()



"""
More specific slurm plugin for checker and submitter,
Extra S is for specific or single (indicates that check job not go through slurm)

"""
import os
from abc import abstractmethod
from uuid import uuid4
from functools import partial

from ..config import conf, history
from ..components import SlurmJob, SlurmTask
from ..utils import flatten_dict
from .slurm import SlurmSub, SlurmChk
from ..components.genfiles import generate_file


class SSlurmSub(SlurmSub):
    def submit_pending(self, jobid):
        sbatch_path = os.path.join(conf["inputs_abs_dir"], jobid + ".sh")
        t = SlurmTask(sbatch_path=sbatch_path)
        t.submit()
        # print(t.jobid())


class SSlurmChk(SlurmChk):
    """
    For subclass to be usable, one need to define methods including:
    _render_input, _render_commands, _render_options_append (if needed)
    _render_resource (default {}, if needed), _render_newid (default uuid1, if needed)
    and specifically check_checking_main
    """

    def __init__(self, params=None, _from="conf", **kwargs):
        """

        :param params:
        :param _from: str. conf, template.
        :param kwargs:
        """
        self._from = _from
        super().__init__(params, **kwargs)

    def _render_input(self, jobid, checkid="", param=None, prefix=""):
        """
        generate input files based on jobid and param.
        The default impl can render param to input.template.
        But this is not general enough for all user case,
        the user can simply rewrite this method.

        :param jobid: str.
        :param param: Unoin[Dict, List, Tuple].
        :return: None.
        """
        info_dict = flatten_dict(
            {"conf": conf, "param": param, "jobid": jobid, "checkid": checkid,}
        )
        if not prefix:
            if conf.get("input_template"):
                generate_file(
                    data=flatten_dict(info_dict),
                    output_path=os.path.join(conf["inputs_abs_dir"], jobid),
                    output_template=os.path.join(
                        conf["work_dir"], conf["input_template"]
                    ),
                )
        elif prefix == "check_":
            if conf.get(prefix + "input_template"):
                generate_file(
                    data=flatten_dict(info_dict),
                    output_path=os.path.join(conf[prefix + "inputs_abs_dir"], checkid),
                    output_template=os.path.join(
                        conf["work_dir"], conf[prefix + "input_template"]
                    ),
                )

    def _render_sbatch(self, jobid, checkid="", param=None, prefix=""):
        if not prefix:
            _sbatch_path = os.path.join(conf["inputs_abs_dir"], jobid + ".sh")
        elif prefix == "check_":
            _sbatch_path = os.path.join(
                conf[prefix + "inputs_abs_dir"], checkid + ".sh"
            )
        if self._from == "conf":
            SlurmTask(
                sbatch_path=_sbatch_path,
                sbatch_commands=self._render_commands(
                    jobid=jobid, checkid=checkid, param=param, prefix=prefix
                ),
            )

        elif self._from == "template":
            _sbatch_template = os.path.join(
                conf["work_dir"], conf[prefix + "slurm_template"]
            )
            info_dict = flatten_dict(
                {"conf": conf, "param": param, "jobid": jobid, "checkid": checkid,}
            )
            generate_file(
                info_dict, output_path=_sbatch_path, output_template=_sbatch_template
            )
            SlurmTask(sbatch_path=_sbatch_path)

        else:
            raise ValueError("_from must be conf or template")

    def _render_commands(self, jobid, checkid="", param=None, prefix=""):
        commands = conf.get(prefix + "slurm_commands", []).copy()
        opts = conf.get(prefix + "slurm_options", []).copy()
        opts = ["#SBATCH " + opt for opt in opts]
        if checkid:
            _id = checkid
        else:
            _id = jobid
        opts.append("#SBATCH --job-name %s" % _id)
        commands = opts + commands
        return self._substitue_opts(
            opts=commands, jobid=jobid, checkid=checkid, param=param
        )

    # def _replace_func(self, jobid, checkid, char):
    #     # a second thought: why not unify with {} as format
    #     if char == "j":
    #         return jobid
    #     elif char == "J":
    #         return checkid
    #     elif char == "i":
    #         return os.path.join(conf.get("inputs_abs_dir", ""), jobid)
    #     elif char == "o":
    #         return os.path.join(conf.get("outputs_abs_dir", ""), jobid)
    #     elif char == "I":
    #         return os.path.join(conf.get("check_inputs_abs_dir", ""), checkid)
    #     elif char == "O":
    #         return os.path.join(conf.get("check_outputs_abs_dir", ""), checkid)
    #     elif char == "e":
    #         return os.path.join(conf["work_dir"], conf.get("executable", ""))
    #     elif char == "E":
    #         return os.path.join(conf["work_dir"], conf.get("check_executable", ""))
    #     elif char == "v":
    #         return conf.get("executable_version", "")
    #     elif char == "V":
    #         return conf.get("check_executable_version", "")
    #     elif char == "w":
    #         return conf.get("work_dir", "")
    #     return ""

    def _substitue_opts(self, opts, jobid, checkid="", param=None):
        """

        :param opts: lits of strings
        :return:
        """
        info_dict = flatten_dict(
            {"conf": conf, "param": param, "jobid": jobid, "checkid": checkid,}
        )
        # print(info_dict)
        for i, opt in enumerate(opts):
            opts[i] = opt.format(**info_dict)  # sep="." doesn't work here
        # f-sring is way better for {a[b]} support naturally, but considering py3.5 here...
        return opts

    def _render_resource(self, jobid, checkid="", param=None):
        res = {}
        res["job_count"] = 1
        return res

    def _render_check(self, params):
        r = []
        for param in params:
            jobid = self._render_newid()
            self._render_input(jobid=jobid, param=param)
            self._render_sbatch(jobid=jobid, param=param)
            r.append((jobid, self._render_resource(jobid=jobid, param=param)))
        return r

    def _render_newid(self):
        return str(uuid4())

    def check_kickstart(self):
        return self._render_check(self.params)

    @abstractmethod
    def check_checking_main(self, jobid):
        """

        :param jobid:
        :return: list of param for new jobs
        """
        return []

    def check_checking(self, jobid):
        params = self.check_checking_main(jobid)
        return self._render_check(params)



from ..framework import PlainChk, PlainSub


class DebugChk(PlainChk):
    def __init__(self, params=None, is_next=False, **kwargs):
        self.is_next = is_next
        super().__init__(params=params, **kwargs)

    def check_checking(self, outputpath):
        print("the output path to be checked is %s" % outputpath)
        if self.is_next is True:
            from uuid import uuid1

            return [(str(uuid1()), {"core": 2})]
        return []


class DebugSub(PlainSub):
    def submit_pending(self, inputpath):
        print("the input path to be submitted as job is %s" % inputpath)
        return



# avoid using pyslurm as it is an external dependence and is not portable as well

import os
import re
import sys
import time
import subprocess
from datetime import datetime

from ..exceptions import SubwayException


slurm_abnormal_states = [
    "BOOT_FAIL",
    "CANCELLED",
    "DEADLINE",
    "FAILED",
    "NODE_FAIL",
    "OUT_OF_MEMORY",
    "PREEMPTED",
    "STOPPED",
    "TIMEOUT",
]


def _slurm_time_trans(timestr, _format="%Y-%m-%dT%H:%M:%S"):
    """
    Fault tolerant time string translate to time obj and time stamps

    :param timestr: str.
    :param _format: Optional[str]. Default as "%Y-%m-%dT%H:%M:%S" used in sacct output
    :return: Tuple[Datetime, float]. Datetime obj and timestamp.
    """
    try:
        timeob = datetime.strptime(timestr, _format)
        timets = timeob.timestamp()
    except ValueError:
        timeob = timestr
        timets = timeob
    return timeob, timets


class SlurmException(SubwayException):
    def __init__(self, message, code=90):
        super().__init__(message, code)


class SlurmValueError(SlurmException):
    def __init__(self, message, code=91):
        super().__init__(message, code)


# TODO: more robust interaction with sacct with various possibilities and error cases
# documentation for sacct: https://slurm.schedmd.com/sacct.html
class SlurmJob:
    """
    Abstraction for job **submitted** to slurm.
    """

    def __init__(self, jobname=None, jobid=None, sacct="sacct"):
        """

        :param jobname: Optional[str].
        :param jobid: Optional[str]. One must specify at least one between ``jobid`` and ``jobname``.
        :param sacct: Optional[str]. Binary for ``sacct`` in shell, default is "sacct".
        """
        self.sacct = sacct
        if (not jobname) and (not jobid):
            raise SlurmValueError("Must specify jobid or jobname")
        if jobid:
            self.jobid = jobid
        else:  # only jobname is defined (it is the user's responsibility to make sure that jobname is unique)
            self.jobid = self.get_jobid(jobname)
        self.jobinfo = self.get_jobinfo(self.jobid)
        self.jobname = self.jobinfo["JobName"]

    def get_jobid(self, jobname, tries=6, interval=0.8):
        """
        get jobid from jobname via sacct query

        :param jobname: str.
        :param tries: Optional[int]. Default 6. It is worthing noting that query soon after job submitted would meet empty line,
                so repetitive try is necessary.
        :param interval: Optional[float]. Default 0.8. Seconds between two tries.
        :return: str. jobid.
        :raises SlurmException: when failed to get jobid after ``tries`` tries.
        """
        for i in range(tries):
            try:
                return self._get_jobid(jobname)
            except SlurmException as e:
                if e.code != 98 or i == tries - 1:
                    raise e
                else:
                    print(e.message, file=sys.stderr)
                    time.sleep(interval)

    def get_jobinfo(self, jobid, tries=6, interval=0.8):
        """
        get job info from jobid via sacct query

        :param jobid: str, jobid in slurm system (not jobid in subway which is jobname in slurm!)
        :param tries: Optional[int]. Default 6. It is worthing noting that query soon after job submitted would meet ilegal line,
                so repetitive try is necessary.
        :param interval:  Optional[float]. Default 0.8. Seconds between two tries.
        :return: Dict. Slurm job info.
        """
        for i in range(tries):
            try:
                return self._get_jobinfo(jobid)
            except AssertionError as e:
                if i == tries - 1:
                    raise e
                else:
                    print("try refetching jobinfo", file=sys.stderr)
                    time.sleep(interval)

    def _get_jobid(self, jobname):
        r = subprocess.run(
            [self.sacct, "--name=%s" % jobname, "--format=JobID%50,Jobname%50"],
            stdout=subprocess.PIPE,
        )
        rl = r.stdout.decode("utf-8").split("\n")
        # print(rl)# seconds after job submit, no line is expected with rl[2] = ""
        if len(rl) > 2 and rl[2]:
            jid = [s for s in rl[2].split(" ") if s][0].strip()
            return jid
        errmsg = "no job name is %s, you may need wait for a second" % jobname
        raise SlurmException(errmsg, code=98)

    def _get_jobinfo(self, jobid):
        """
        get job relavant info from sacct by jobid

        :param jobid: str.
        :return: jobinfo: Dict[str, str], ``{'User': 'linuxuser', 'JobID': '4500', 'JobName': 'uuid',
                                'Partition': 'general', 'State': 'COMPLETED', 'Timelimit': '365-00:00+',
                                'Start': '2020-02-23T10:05:55', 'End': '2020-02-23T10:06:15',
                                'Elapsed': '00:00:20', 'NNodes': '1', 'NCPUS': '2'}``
        """
        # nodelist can be "None assigned"
        # Timelimit attr can also be problematic in some slurm, the result is nothing
        r = subprocess.run(
            [
                self.sacct,
                "-j",
                jobid,
                "--format=User%30,JobID%50,Jobname%50,partition%20,state%20,time,start,end,elapsed,nnodes,ncpus,nodelist",
                "-P",  # that is the key point for accurate parsing!
            ],
            stdout=subprocess.PIPE,
        )
        rl = r.stdout.decode("utf-8").split("\n")
        rl = rl[:2]
        rll = [[s.strip() for s in l.split("|")] for l in rl]

        assert len(rll[0]) == len(rll[1])
        info = {}
        for i, head in enumerate(rll[0]):
            info[head] = rll[1][i]
        info["Start_ob"], info["Start_ts"] = _slurm_time_trans(info["Start"])
        if info.get("End", ""):
            info["End_ob"], info["End_ts"] = _slurm_time_trans(info["End"])
        return info


class SlurmTask:
    """
    Abstraction for slurm job **to be submitted**.
    """

    def __init__(
        self,
        sbatch="sbatch",
        scancel="scancel",
        shebang="#!/bin/bash",
        sbatch_path=None,
        sbatch_options=None,
        sbatch_commands=None,
    ):
        """

        :param sbatch: string, binary for sbatch
        :param sbatch: string, binary for scancel
        :param shebang: string, the #! line
        :param sbatch_path: string, sbatch script path
        :param sbatch_options: list of strings, such as "--job-name=uuid"
        :param sbatch_commands: list of strings, main command, such as "python test.py"
        """
        self.sbatch = sbatch
        self.scancel = scancel
        if not sbatch_path:
            raise SlurmValueError("sbatch_path must be specified")
        self.sbatch_path = sbatch_path
        self.shebang = shebang
        if not sbatch_commands:
            sbatch_commands = []
        self.sbatch_commands = sbatch_commands
        if not sbatch_options:
            sbatch_options = []
            # TODO: support dict sbatch options, may a consistent option API for subway?
        self.sbatch_options = sbatch_options
        if not os.path.exists(sbatch_path):
            self._render_sbatch()
        self.jid = None
        self._slurm_outpath = None

    def _render_sbatch(self):
        """
        generate sbatch script for the task.

        :return: None.
        """
        sbatch_string = self.shebang + "\n"
        for opt in self.sbatch_options:
            sbatch_string += "#SBATCH  " + opt + "\n"
        for line in self.sbatch_commands:
            sbatch_string += line + "\n"
        sbatch_string += "\n"
        with open(self.sbatch_path, "w") as f:
            f.writelines([sbatch_string])
        os.chmod(self.sbatch_path, 0o700)

    def slurm_outpath(self):
        # very experimental, not recommended for any production environment
        # array job default out: slurm-%A_%a.out, plain job default: slurm-%j.out
        # -e --error
        # -o --output
        # TODO: support customization on output file
        # TODO: support jobname pattern?
        # TODO: support on array job output
        # TODO: relative output path
        outname = []
        if not self._slurm_outpath:
            # for opt in self.sbatch_options:
            #     if opt.startswith(("-e", "--error", "-o", "--output")):
            #         outname.append()
            _slurm_outpath = os.path.join(os.getcwd(), "slurm-" + self.jobid() + ".out")
            self._slurm_outpath = [_slurm_outpath]
        return self._slurm_outpath

    def submit(self):
        """
        submit jobs to slurm by ``sbatch`` in shell.

        :return: None.
        """
        if not os.path.exists(self.sbatch_path):
            raise SlurmException("No sbatch file at %s" % self.sbatch_path, code=92)
        self.outerr = subprocess.run(
            [self.sbatch, self.sbatch_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    def stdouterr(self):
        """
        get stdout and stderr for sbatch command

        :return: Tuple[str, str], string for stdout and stderr.
        """
        stdout = self.outerr.stdout
        stderr = self.outerr.stderr
        if stdout:
            stdout = stdout.decode("utf-8")
        if stderr:
            stderr = stderr.decode("utf-8")
        return stdout, stderr

    def jobid(self):
        """
        Get jobid from stdout of ``sbatch``.

        :return: str. jobid
        :raises SlurmException: No legal stdout from sbatch command to capture the jobid.
        """
        if not self.jid:
            stdout, stderr = self.stdouterr()
            if stdout:
                l = re.search(r"Submitted batch job (\d.*)", stdout)
                self.jid = l.groups()[0]
                return self.jid
            raise SlurmException(
                "No stdout for sbatch submit, the err is %s" % stderr, code=93
            )
        else:
            return self.jid

    def cancel(self):
        """
        cancel the job

        :return: subprocess.CompletedProcess, from ``scancel``.
        """
        self.jobid()
        r = subprocess.run(
            [self.scancel, self.jid], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        return r

    def delete(self, include_output=False):
        """
        Delete the sbatch script and slurm output (experimental)

        :param include_output: Optional[bool]. Default False, whether try to delete slurm output
        :return: None.
        """
        if os.path.exists(self.sbatch_path):
            os.remove(self.sbatch_path)
        else:
            print("sbatch file doesn't exist", file=sys.stderr)
        if include_output:
            for f in self.slurm_outpath():
                if os.path.exists(f):
                    os.remove(f)
                else:
                    print("output file %s is not generated" % f, file=sys.stderr)



"""
Provide DSlurmSub and DSlurmChk class, with consideration on check scripts into slurm system
"""

import os
from abc import abstractmethod

from ..config import conf, history
from ..components import SlurmJob, SlurmTask
from .sslurm import SSlurmSub, SSlurmChk


class DSlurmSub(SSlurmSub):
    def submit_aborted(self, jobid):
        sbatch_path = os.path.join(
            conf["work_dir"], conf["check_inputs_dir"], history[jobid]["assoc"] + ".sh"
        )
        t = SlurmTask(sbatch_path=sbatch_path)
        t.submit()

    def submit_finished(self, jobid):
        self.submit_aborted(jobid)


# TODO: sslurm is expected to merge into dslurm totally
class DSlurmChk(SSlurmChk):
    """
    For subclass to be usable, one need to define methods including:
    _render_input, _render_commands, _render_options_append (if needed)
    _render_resource (default {}, if needed), _render_newid (default uuid1, if needed)
    and specifically check_checking_main
    """

    def _render_check(self, params, jobid=None, _type="main"):
        if _type == "main":
            r = []
            for param in params:
                jobid = self._render_newid()
                self._render_input(jobid=jobid, param=param)
                self._render_sbatch(jobid=jobid, param=param)
                r.append((jobid, self._render_resource(jobid=jobid, param=param)))
            return r
        elif _type == "check":
            if params:
                assert len(params) == 1
                param = params[0]
                checkid = self._render_newid()
                self._render_input(
                    jobid=jobid, checkid=checkid, param=param, prefix="check_"
                )
                self._render_sbatch(
                    jobid=jobid, checkid=checkid, param=param, prefix="check_"
                )
                return [
                    (
                        checkid,
                        self._render_resource(
                            jobid=jobid, checkid=checkid, param=param
                        ),
                    )
                ]

    def check_finished(self, jobid):  # should generate check input and check sbatch
        params = self.check_finished_main(jobid)
        return self._render_check(params=params, jobid=jobid, _type="check")

    @abstractmethod
    def check_finished_main(self, jobid):
        """

        :param jobid: str.
        :return: List[Tuple[str, Dict[str, Any]]]. The length of the list must be 0 or 1.
                ``[(checkid, check_resource)]``
        """
        return []



"""
cli interface for sub init and more
"""

import argparse  # click might be better, but as hpc app, we try our best to only use builtin module
import os
import sys
import json
import shutil
from functools import partial
from datetime import datetime

from .bootstrap import env_init
from .utils import (
    load_json,
    editor,
    print_json,
    ts2str,
    simple_template_render,
    statement_parser,
)
from .htree import HTree
from .exceptions import CLIException, NoAttribute
from . import version


class SubwayCLI:
    def __init__(self, _argv=None):
        parser = argparse.ArgumentParser(description="subway is comming...")
        parser.add_argument(
            "-d", "--directory", dest="dir", default=os.getcwd(), help="project dir"
        )
        parser.add_argument(
            "-V",
            "--version",
            dest="version",
            default=False,
            action="store_true",
            help="subway version",
        )
        subparsers = parser.add_subparsers(dest="command")
        # init params
        initparser = subparsers.add_parser(
            "init", aliases="i", description="init subway project"
        )
        initparser.add_argument(
            "-c", "--config", dest="conf", default=None, help="config.json path"
        )
        # help params
        helpparser = subparsers.add_parser(
            "help", aliases="h", description="help on subway"
        )
        # run params
        runparser = subparsers.add_parser(
            "run", aliases="r", description="check and submit jobs"
        )
        # query params
        queryparser = subparsers.add_parser(
            "query", aliases="q", description="query information on jobs"
        )
        queryparser.add_argument(
            "-j", "--job", dest="path", default=None, help="the task investigated"
        )
        queryparser.add_argument(
            "-s", "--statement", dest="statement", default=None, help="the query string"
        )
        queryparser.add_argument(dest="action", help="query action for task", nargs="?")
        configparser = subparsers.add_parser(
            "config", aliases="c", description="config subway"
        )
        configparser.add_argument(dest="action", help="action on config")
        debugparser = subparsers.add_parser(
            "debug", aliases="d", description="debug related commands"
        )
        debugparser.add_argument(dest="object", help="object for debug")
        debugparser.add_argument(
            dest="action", help="action for debug", nargs="?", default=None
        )
        debugparser.add_argument("-f", "--from", dest="_from", default="template")
        # TODO: automatic way to add help for subparser, eg. decorator of query funcs
        # TODO: plugable subcommands, eg slurm
        self.parser = parser
        if not _argv:
            _argv = sys.argv[1:]
        self.args = parser.parse_args(_argv)
        # print(_argv, self.args)
        self.tree = None
        self._subway_path = os.path.dirname(__file__)

        for a in [
            "reason",
            "assoc",
            "prev",
            "next",
            "resource",
            "executable_version",
            "check_executable_version",
        ]:
            self._meta_query_attr(a)

        for k in [
            "pending",
            "running",
            "finished",
            "aborted",
            "checking",
            "resolving",
            "checked",
            "frustrated",
            "failed",
            "resolved",
        ]:
            self._meta_query_key(k)

        for t in [
            "creating",
            "beginning",
            "finishing",
            "checking",
            "ending",
            "beginning_real",
        ]:
            self._meta_query_ts(t)

    def __call__(self):
        try:
            if self.args.version:
                print("subway version:", version)
                return
            if self.args.command not in ["init", "i", "help", "h", "debug", "d"]:
                # there is no system independent elegant way to get root path in python
                # see https://stackoverflow.com/questions/17429044/constructing-absolute-path-with-os-path-join
                cp = self.args.dir
                # last path, current path
                while True:
                    if os.path.exists(os.path.join(cp, ".subway")):
                        self.dir = cp
                        break
                    lp = cp
                    cp = os.path.dirname(cp)
                    if cp == lp:  # confirm lp is root path
                        raise CLIException(
                            message="fatal: Not a subway project", code=14
                        )

                self.conf = load_json(os.path.join(self.dir, ".subway", "config.json"))
                self.history = load_json(
                    os.path.join(self.dir, ".subway", "history.json")
                )
                # conf cannot load for init

            getattr(self, self.args.command, "help")()
        except CLIException as e:
            print(e.message, file=sys.stderr)
            exit(e.code)

    def init(self):
        if self.args.conf is not None:
            with open(self.args.conf, "r") as f:
                conf = json.load(f)
        else:
            conf = None

        env_init(self.args.dir, conf)

        print(
            "the subway project has been successfully initialized in %s" % self.args.dir
        )

    i = init

    def run(self):

        executer = self.conf.get("entry_point", "main.py")
        os.system(os.path.join(self.dir, executer))

    r = run

    def query(self):
        if not self.tree:
            self.tree = HTree(self.history)
        self.jid = self.args.path
        if self.jid:
            self.jid = self.tree.match(self.jid)
            # print(self.jid)
            print("matched job id is %s" % self.jid, file=sys.stderr)
        if self.args.statement:
            self.args.action = "condition"
        if not self.args.action:
            raise CLIException("Must specify one action for query")
        return getattr(self, "query_" + self.args.action, self.query_info)()

    q = query
    # beginning_real_ts, reason, assoc

    def query_info(self):
        if self.jid:
            print_json(self.history[self.jid])
        else:
            print_json(self.history)

    def query_tree(self):
        if self.jid:
            self.tree.print_tree(self.jid)
        else:
            self.tree.print_trees(self.tree.roots())

    query_t = query_tree

    def query_input(self):
        if self.jid:
            inputpath = os.path.join(
                self.dir, self.conf.get("inputs_dir", ""), self.jid
            )
            if os.path.exists(inputpath):
                with open(inputpath, "r") as f:
                    content = f.readlines()
                print("".join(content))
            else:
                raise CLIException("no input files for %s" % self.jid)
        else:
            self._noid()

    query_i = query_input

    def query_output(self):
        if self.jid:
            outputpath = os.path.join(
                self.dir, self.conf.get("outputs_dir", ""), self.jid
            )
            if os.path.exists(outputpath):
                with open(outputpath, "r") as f:
                    content = f.readlines()
                print("".join(content))
            else:
                raise CLIException("no output files for %s" % self.jid)
        else:
            self._noid()

    query_o = query_output

    def query_root(self):
        if self.jid:
            print(self.tree.root(self.jid))
        else:
            print(*self.tree.roots(), sep="\n")

    query_r = query_root

    def query_leaves(self):
        if self.jid:
            print(*self.tree.end(self.jid), sep="\n")
        else:
            print(*self.tree.leaves(), sep="\n")

    query_l = query_leaves

    def query_state(self):
        if self.jid:
            print(self.history[self.jid]["state"])
        else:
            s = set()
            for _, st in self.history.items():
                s.add(st["state"])
            print(*s, sep="\n")

    query_s = query_state

    @staticmethod
    def _get(s, k, _default=None):
        """
        ``_get({"a": {"b":1}}, "a.b") = 1``

        :param s:
        :param k:
        :param _default:
        :return:
        """
        ks = k.split(".")
        d = s
        for i in ks:
            if d.get(i):
                d = d[i]
            else:
                return _default
        return d

    @staticmethod
    def _set(s, k, v):
        """
        ``_set({"a":{"b":{"c":2}, "zz":"pp"}}, "a.b.m", "q")={'a': {'b': {'c': 2, 'm': 'q'}, 'zz': 'pp'}}``

        :param s: Dict.
        :param k: Any.
        :param v: Any.
        :return: Dict. after the set
        """
        ks = k.split(".")
        d = s
        for c, i in enumerate(ks):
            if not isinstance(d.get(i), dict):
                if c == len(ks) - 1:
                    d[i] = v
                    break
                else:
                    return s
            d.setdefault(i, {})
            d = d[i]

        return s

    def query_condition(self):
        # print(self.args.statement) #debug,
        try:
            st = statement_parser(self.args.statement)
        except (ValueError, TypeError) as e:
            raise CLIException(message=e.args[0])
        # print(st)
        joblist = []
        tskset = set()  # {(beginning_tso, beginning_ts)}
        for k, v in st.items():
            if k.endswith("_tso"):
                # tsk = "".join(k.split("_")[:-1]) + "_ts"
                tsk = k[:-1]
                tskset.add((k, tsk))
        for tso, tsk in tskset:
            for j, s in self.history.items():
                if s.get(tsk):
                    s[tso] = datetime.fromtimestamp(s[tsk])
        for j, s in self.history.items():
            for k, v in st.items():

                if v[0] == "=":
                    if self._get(s, k):
                        # print(j, k, s[k], v[1])
                        if self._get(s, k) != v[1]:
                            break
                    else:
                        break
                elif v[0] == "<":
                    if self._get(s, k, "") != "":
                        if (
                            isinstance(v[1], list)
                            or isinstance(v[1], set)
                            or isinstance(v[1], tuple)
                            or isinstance(v[1], dict)
                        ):
                            if self._get(s, k, "") not in v[1]:
                                break
                        elif self._get(s, k, "") >= v[1]:
                            break
                    else:
                        break
                elif v[0] == ">":
                    if self._get(s, k, "") != "":  # s.get(k) is not enough,
                        # since we also have to exclude [] namely [] should return true

                        if (
                            isinstance(self._get(s, k), list)
                            or isinstance(self._get(s, k), set)
                            or isinstance(self._get(s, k), tuple)
                            or isinstance(self._get(s, k), dict)
                        ):
                            if v[1] not in self._get(s, k, ""):
                                break
                        elif self._get(s, k, "") <= v[1]:
                            break
                    else:
                        break
                elif v[0] == "<=":
                    if self._get(s, k):
                        if self._get(s, k) > v[1]:
                            break
                    else:
                        break
                elif v[0] == ">=":
                    if self._get(s, k):
                        if self._get(s, k) < v[1]:
                            break
                    else:
                        break
                elif v[0] == "<>":  # unequal sign
                    if self._get(s, k):
                        if self._get(s, k) == v[1]:
                            break
                    else:
                        break
            else:
                joblist.append(j)
        print(*joblist, sep="\n")

    def _meta_query_attr(self, key):
        def f(s):
            if s.jid:
                v = self.history[self.jid].get(key, "")
                if isinstance(v, list):
                    print(*v, sep="\n")
                else:
                    print(v)
            else:
                s._noid()

        setattr(self, "query_" + key, partial(f, self))

    def _meta_query_key(self, key):
        """
        meta function to generate member function for this class, generate function as follows

        .. code-block:: python

            def query_checked(self):
                if self.jid:
                    print("1" if self.history[self.jid]['state'] == "checked" else "0")
                else:
                    r = []
                    for j, s in self.history.items():
                        if s['state'] == "checked":
                            r.append(j)
                    print(*r, sep="\\n")


        :param key: str. The added function will be named as ``query_key``
        :return: None
        """

        def f(s):
            if s.jid:
                print("1" if self.history[self.jid]["state"] == key else "0")
            else:
                r = []
                for j, d in s.history.items():
                    if d["state"] == key:
                        r.append(j)
                print(*r, sep="\n")

        setattr(self, "query_" + key, partial(f, self))

    def _meta_query_ts(self, tstype):
        def f(s, readable=True):
            if s.jid:
                ts = self.history[s.jid].get(tstype + "_ts", "")
                if ts:
                    if readable:
                        ts = ts2str(ts)
                    print(ts)
                else:
                    raise NoAttribute("no %s timestamps for %s" % (tstype, self.jid))
            else:
                self._noid()

        pfstr = partial(f, self, True)
        pfts = partial(f, self, False)
        # setattr(self, "query_" + tstype, pfstr) # may be conflict to state search
        setattr(self, "query_" + tstype + "_ts", pfts)
        setattr(self, "query_" + tstype + "_time", pfstr)

    def config(self):
        if self.args.action == "show" or self.args.action == "s":
            print_json(self.conf)
        elif self.args.action == "edit" or self.args.action == "e":
            editor(os.path.join(self.dir, ".subway", "config.json"))

    c = config

    def help(self):
        self.parser.print_help()

    h = help

    def _history_clean(self):
        with open(os.path.join(self.args.dir, ".subway", "history.json"), "w") as f:
            json.dump({}, f)

    def _render_run_check(self, runpath, checkpath=None, subwaypath=None):
        """

        :param runpath: relative path (only file name)
        :param checkpath:
        :param subwaypath: dir path for subway lib
        :return:
        """
        if not subwaypath:
            subwaypath = self._subway_path
        if checkpath:
            shutil.copyfile(
                os.path.join(subwaypath, "examples", "miscs", checkpath),
                os.path.join(self.args.dir, "check.py"),
            )
        shutil.copyfile(
            os.path.join(subwaypath, "examples", "miscs", runpath),
            os.path.join(self.args.dir, "run.py"),
        )

    def _render_sbatch_template(self, mainpath, checkpath=None, subwaypath=None):
        if not subwaypath:
            subwaypath = self._subway_path
        if checkpath:
            shutil.copyfile(
                os.path.join(subwaypath, "examples", "miscs", checkpath),
                os.path.join(self.args.dir, "check_sbatch.template"),
            )
        shutil.copyfile(
            os.path.join(subwaypath, "examples", "miscs", mainpath),
            os.path.join(self.args.dir, "sbatch.template"),
        )

    def debug(self):
        if self.args.object == "history":
            if self.args.action in ["clear", "clean"]:
                self._history_clean()

        if self.args.object == "reinit":
            self._history_clean()
            self.conf = load_json(os.path.join(self.args.dir, ".subway", "config.json"))
            for p in [
                os.path.join(self.args.dir, self.conf["inputs_dir"]),
                os.path.join(self.args.dir, self.conf["outputs_dir"]),
                os.path.join(
                    self.args.dir, self.conf.get("check_inputs_dir", "cinputs")
                ),
                os.path.join(
                    self.args.dir, self.conf.get("check_outputs_dir", "coutputs")
                ),
            ]:
                if os.path.exists(p):
                    shutil.rmtree(p)
                os.mkdir(p)

        if self.args.object == "setup":
            if self.args.action == "rgd":
                conf = load_json(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_config.json"
                    )
                )
                conf["_py"] = sys.executable
                env_init(self.args.dir, conf, include_main=False)
                var_dict = {
                    "_py": sys.executable,
                    "_sub": "RgDSub",
                    "_chk": "RgDChk",
                    "_from": self.args._from,
                }
                simple_template_render(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_main.template"
                    ),
                    os.path.join(self.args.dir, conf.get("entry_point", "main.py")),
                    var_dict,
                )
                self._render_run_check("rg_run.py", "rg_check.py")
                self._render_sbatch_template(
                    "rg_sbatch.template", "rg_check_sbatch.template"
                )
                shutil.copy(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_input.template"
                    ),
                    os.path.join(self.args.dir, "input.template"),
                )

            elif self.args.action == "rgs":
                conf = load_json(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_config.json"
                    )
                )
                conf["_py"] = sys.executable
                env_init(self.args.dir, conf, include_main=False)
                var_dict = {
                    "_py": sys.executable,
                    "_sub": "RgSSub",
                    "_chk": "RgSChk",
                    "_from": self.args._from,
                }
                simple_template_render(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_main.template"
                    ),
                    os.path.join(self.args.dir, conf.get("entry_point", "main.py")),
                    var_dict,
                )
                self._render_run_check("rg_run.py")
                self._render_sbatch_template("rg_sbatch.template")
                shutil.copy(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_input.template"
                    ),
                    os.path.join(self.args.dir, "input.template"),
                )

            elif self.args.action == "rgl":
                conf = load_json(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_config.json"
                    )
                )
                conf["_py"] = sys.executable
                env_init(self.args.dir, conf, include_main=False)
                var_dict = {
                    "_py": sys.executable,
                    "_sub": "RgSub",
                    "_chk": "RgChk",
                    "_from": self.args._from,
                }
                simple_template_render(
                    os.path.join(
                        self._subway_path, "examples", "miscs", "rg_main.template"
                    ),
                    os.path.join(self.args.dir, conf.get("entry_point", "main.py")),
                    var_dict,
                )
                self._render_run_check("rg_run.py")
            os.chmod(
                os.path.join(self.args.dir, conf.get("entry_point", "main.py")), 0o700
            )

    d = debug

    def _noid(self):
        raise CLIException("Please specify job id", code=12)



"""
Forest of multiple trees of tasks
"""

from .exceptions import MatchError


class HTree:
    """
    Trees structure for tasks recorded in history.json.
    """

    def __init__(self, history):
        self.history = history
        self.roots_list = None
        self.leaves_list = None

    def match(self, idprefix):
        results = []
        for j, _ in self.history.items():
            if j.startswith(idprefix):
                results.append(j)
        if not results:
            raise MatchError("Job id prefix matches no job")
        if len(results) > 1:
            raise MatchError("Job id prefix matches multiple jobs")
        return results[0]

    def parent(self, jobid, n=1):
        """
        n-level parent task id

        :param jobid:
        :param n: n<=0 for the ultimate root
        :return:
        """
        p1 = self.history[jobid]["prev"]
        # print(p1)
        if p1 is None:
            if n > 0:
                return
                # raise ValueError("%s has no parent in that level" % jobid)
            else:
                return jobid
        if n == 1:
            return p1
        return self.parent(p1, n - 1)

    def roots(self):
        """
        Find all jobs with no parent.

        :return: List[str]. jobid list
        """
        if self.roots_list is not None:
            return self.roots_list
        r = []
        for jid, s in self.history.items():
            if s["prev"] is None:
                r.append(jid)
        self.roots_list = r
        return r

    def root(self, jobid):
        return self.parent(jobid, n=0)

    def children(self, jobid, n=1):
        """

        :param jobid: jobid or jobids list
        :param n: n=0 for the lowest level
        :return: list of children jobid
        """
        c1s = []
        if type(jobid) is not list:
            jobid = [jobid]
        for j in jobid:
            c1 = self.history[j]["next"]
            if c1:
                c1s.extend(c1)
        if n == 1:
            return c1s
        if not c1s and n <= 0:
            return jobid
        return self.children(c1s, n - 1)

    def end(self, jobid):
        """
        downstream of jobid as leaves

        :param jobid:
        :return: list
        """
        r = []
        for node in self.DFSvisit(jobid):
            if not self.history[node]["next"]:
                r.append(node)
        return r

    def DFSvisit(self, jobid):
        """
        generator yield job by DFS across the tree rooted as ``jobid``

        :param jobid: str.
        :yield: Next job id by DFS.
        """
        yield jobid
        for nid in self.history[jobid]["next"]:
            yield from self.DFSvisit(nid)

    def BFSvisit(self, jobid):
        pass  # TODO: BFS

    def leaves(self):
        if self.leaves_list is not None:
            return self.leaves_list
        l = []
        for jid, s in self.history.items():
            if not s["next"]:
                l.append(jid)
        self.leaves_list = l
        return l

    def print_tree(self, jobid, file=None, _prefix="", _last=True, _show=0):
        """
        print a tree to stdout with root as ``jobid``

        :param jobid: str.
        :param file: args for ``file=`` in ``print``. Default None to stdout.
        :param _prefix:
        :param _last:
        :param _show:
        :return:
        """
        print(
            _prefix,
            "`- " if _last else "|- ",
            jobid if _show == 0 else jobid[:_show],
            sep="",
            file=file,
        )
        _prefix += "   " if _last else "|  "
        children = self.history[jobid]["next"]
        child_count = len(children)
        for i, child in enumerate(children):
            _last = i == (child_count - 1)
            self.print_tree(child, file, _prefix, _last, _show)

    def print_trees(self, jobids, file=None, _prefix="", _last=True, _show=0):
        for jobid in jobids:
            self.print_tree(jobid, file, _prefix, _last, _show)



"""
smarter file rendered can be utilized in _render_input of SSlurmChk
"""
import json

from ..utils import simple_template_render, flatten_dict

# TODO: provide smart unified API to generate files with given python data structure
# .xml .json .yaml/toml .txt .csv
def generate_file(
    data, output_path, output_format=None, output_config=None, output_template=None
):
    if not output_config:
        output_config = {}

    # firstly transform everything to standard intermediate state as a dict json like
    data = jsonify(data, _outer_most=True)
    if output_template:
        simple_template_render(output_template, output_path, flatten_dict(data))
    if output_format == "json":
        with open(output_path, "w") as f:
            json.dump(data, f)
            return


def jsonify(data, _outer_most=False, _outer=True, func=None):
    if data is None:
        return None
    elif isinstance(data, dict):  # auto support collections.OrderedDict
        for k, v in data.items():
            data[k] = jsonify(v, _outer=False, func=func)
    elif isinstance(data, list) or isinstance(data, set) or isinstance(data, tuple):
        data = list(data)
        for i, d in enumerate(data):
            data[i] = jsonify(d, _outer=True, func=func)
        if _outer:
            data = {"list": data}
    elif isinstance(data, str) or isinstance(data, int) or isinstance(data, float):
        if _outer_most:
            data = {type(data).__name__: data}
    else:
        if func:
            data = func(data)
        else:
            raise ValueError("Unsupported data type %s to jsonify" % type(data))

    return data



"""
some utilities and helper function with no inner dependence
"""

import os
import re
import sys
import json
import shlex
import hashlib
import subprocess
from functools import partial
from datetime import datetime


def now_ts():
    """
    get timestamps of now

    :return: float, timestamps
    """
    return datetime.now().timestamp()


def ts2str(ts, _format="%Y-%m-%d, %H:%M:%S"):
    """
    transform timestamp to corresponding formatted string

    :param ts: int, timestamps.
    :return: str, eg. "2020-02-02, 02:02:20"
    """
    do = datetime.fromtimestamp(ts)
    return do.strftime(_format)


def load_json(conf_file):
    """
    load json file and return dict

    :param conf_file: str, absolute json file path
    :return: Dict.
    """
    with open(conf_file, "r") as fp:
        conf_dict = json.load(fp)
    return conf_dict


def print_json(json_dict, indent=2):
    """
    print python dict with pretty indent

    :param json_dict: Dict.
    :param indent: Optional[int], default 2.
    :return: None.
    """
    print(json.dumps(json_dict, indent=indent))


def md5file(path):
    """
    return md5 for given file

    :param path: str, abspath for the file.
    :return: str. md5 value
    """
    BUF_SIZE = 65536  # memory efficient impl
    md5 = hashlib.md5()

    with open(path, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)

    return md5.hexdigest()


def _flatten_tolist(d, parent_key, sep):
    items = []
    if isinstance(d, list) or isinstance(d, tuple):
        for i, j in enumerate(d):
            items.extend(_flatten_tolist(j, "%slist_%s%s" % (parent_key, i, sep), sep))
    elif isinstance(d, dict):
        for k, v in d.items():
            items.extend(_flatten_tolist(v, "%s%s%s" % (parent_key, k, sep), sep))
    else:  # d is simply a number or str
        items.extend([("%s" % (parent_key[:-1],), d)])
    return items


def flatten_dict(d, parent_key="", sep="~"):
    """
    Flatten a nested dict with list. eg. transform ``{"a":{"b":"c", "d": ["e", "f"]}}`` to
    ``{"a.b": c, "a.d.list_0": "e" ...}``

    :param d: Dict[str, Any]. Nested dict with possibly nested list and tuple.
    :param parent_key: Optional[str]. Default ``""``. The common key prefix.
    :param sep: Optional[str], default ``"~"``. The separator between different level of keys.
    :return: Dict[str, Any]. Flattened dict.
    """
    return dict(_flatten_tolist(d, parent_key, sep))


def simple_template_render(template, output, var_dict):
    """
    render template file to output file
    with variable substitution defined by var_dict.
    template file vars are defined within {var}.

    :param template: str. Absolute file name for template file.
    :param output: str. Absolute file name for output file (create or overwrite).
    :param var_dict: Dict[str, str].
    :return: None.
    """
    # jinja is of course better but again no external dependence is required
    with open(template, "r") as f:
        l = f.read()
    nl = l.format(**var_dict)
    with open(output, "w") as f:
        f.writelines([nl])


def statement_parser(st):
    """
    Parse "aa=b; c>13;e<f g>=h" into {"aa": ("=", "b"), "c": (">", 13)...}

    :param st: str.
    :return: Dict[str, Tuple[str, Any]].
    """
    l = re.split(r"[;\s]", st)
    r = {}
    pattern = re.compile(r"[^=><]*([=><]{1,2})[^=><]*")
    for s in l:
        if s:
            sep = pattern.match(s)
            if not sep:
                raise ValueError("illegal query statement")
            sep = sep.groups()[0]
            b, a = s.split(sep)
            a = _recover_type(a)
            b = _recover_type(b)
            r[b] = (sep, a)
    return r


def _recover_type(v):
    """
    recover "17" to 17.
    recover "[]" to [].
    recover "datetime(2020,1,1)" to datetime(2020,1,1).
    keep "abc" as str.

    :param v: str.
    :return: Any.
    """
    try:
        nv = eval(v)
    except (NameError, SyntaxError):
        nv = v
    if callable(nv):
        nv = v
    return nv


def _replace(replace_func, s):
    """
    inner function for :py:func:`replace_wildcard`

    :param replace_func: Callable[[char], char].
        Char to char function. eg f(a) = A, indicates %a should be replaced by A.
    :param s: str, that match r"[\%]+[^\%]*"
    :return: str, transformed after wildcard substitution
    """

    if len(s) <= 1:
        return s
    if s[0] != "%":
        return s
    if s[-1] == "%":
        return s
    if s[1] == "%":
        return s[1:]
    if s[1] != "%":
        return replace_func(s[1]) + s[2:]


def replace_wildcard(replace_func, s):
    """
    replace %a type wildcard in the string based on customized function replace_func.

    :param replace_func: Callable[[char], char].
        Char to char function. eg f(a) = A, indicates %a should be replaced by A.
    :param s: str, the string possibly with %a type wildcards, eg. "bc%de%%a"
    :return: str, the string with %a type wildcards all replaced
    """
    # this utility is not used in latest codebase anymore.
    state = 0
    # 0: the last one in normal 1: the last one is %
    rl = []
    r = ""
    for c in s:
        if c == "%" and state == 0:
            rl.append(r)
            r = c
            state = 1
        elif c == "%" and state == 1:
            r += c
            state = 1
        else:
            r += c
            state = 0
    rl.append(r)
    rl = [r for r in rl if r]
    _preplace = partial(_replace, replace_func)
    return "".join(list(map(_preplace, rl)))


# editors to try if VISUAL and EDITOR are not set
_default_editors = ["vim", "vi", "emacs", "nano"]


def which(binary, path=None):
    """
    Like `which` in linux.
    Find absolute binary path given command shortcut binary.

    :param binary: str. name for binary command.
    :param path: Optional[str], default None. If None, `PATH` is searched
    :return: Optional[str], absolute path for binary.
        None if binary not found in path.
    """
    if not path:
        path = os.environ.get("PATH", "")

    if isinstance(path, str):
        path = path.split(os.pathsep)

    for directory in path:
        exe = os.path.join(directory, binary)
        if os.path.isfile(exe) and os.access(exe, os.X_OK):
            return exe

    return None


def editor(path, **kwargs):
    """
    Invoke the user's editor.
    This will try to execute the following, in order:
    1. $VISUAL <args>    # the "visual" editor (per POSIX)
    2. $EDITOR <args>    # the regular editor (per POSIX)
    3. some default editor (see ``_default_editors``) with <args>
    If an environment variable isn't defined, it is skipped.  If it
    points to something that can't be executed, we'll print a
    warning. And if we can't find anything that can be executed after
    searching the full list above, we'll raise an error.

    :param path: str. Absolute path for file to be opened by editor.
    :return: bool. True if editor is found and open.
    """
    # allow this to be customized for testing
    _run = kwargs.get("_run", subprocess.run)

    def try_exec(exe):
        """
        Try to execute an editor with ``subprocess.run``, and warn if it fails.

        :return: (bool) False if the editor failed, ideally does not
            return if ``run`` succeeds, and ``True`` if the
            ``run`` does return successfully.
        """
        try:
            _run(exe)
            return True

        except OSError as e:
            # Show variable we were trying to use, if it's from one
            print("Could not execute %s due to error:" % exe, str(e), file=sys.stderr)
            return False

    def try_env_var(var):
        """
        Find an editor from an environment variable and try to exec it.
        This will warn if the variable points to something is not
        executable, or if there is an error when trying to exec it.
        """
        if var not in os.environ:
            return False

        return try_exec(shlex.split(os.environ[var] + " " + path))

    # try standard environment variables
    if try_env_var("VISUAL"):
        return
    if try_env_var("EDITOR"):
        return

    # nothing worked -- try the default we can find
    for exe in _default_editors:
        if try_exec(shlex.split(which(exe) + " " + path)):
            return

    # Fail if nothing could be found
    print(
        "No text editor found! Please set the VISUAL and/or EDITOR "
        "environment variable(s) to your preferred text editor."
    )



"""
actions when sub init in cli, including mkdir, touch files and render templates in templates fold
"""

import os
import sys
import json
import shutil

## note conf here is not from config.json, since there is no such file at the phase
## the source of the conf here should be interactive cli or specify json file from sub init -f config.json
def env_init(path, conf=None, fromfile=None, include_main=True):
    """
    initialize a subway project including:
    mkdir .subway, and inputs/outputs check inputs/outputs dirs if possible
    create config.json and empty history.json within .subway
    render an example main.py

    :param path: str. absolute path for subway project dir
    :param conf: Optional[Dict], Default None, indicates a simple default config is applied.
    :param fromfile: Optional[str]. Default None, indicates a default example main.py will be rendered.
                If specified as a full path, the corresponding file will be copied into subway project
                as entry point.
    :param include_main: Optional[bool], default True. If False, main.py is not rendered.
    :return: None.
    """
    if not conf:
        conf = default_conf(path)
    mkdirs(path, conf)
    render_config(path, conf)
    render_history(path)
    if include_main:
        render_main(path, conf, fromfile)


def default_conf(path):
    """
    generate default config dict for subway project

    :param path: str. ``conf["work_dir] = path``
    :return: Dict. The default config dict.
    """
    conf = {}
    conf["inputs_dir"] = "inputs"
    conf["outputs_dir"] = "outputs"
    conf["check_inputs_dir"] = "cinputs"
    conf["check_outputs_dir"] = "coutputs"
    conf["work_dir"] = path  # not necessary, can serve as a double check
    conf["resource_limit"] = {}
    conf["entry_point"] = "main.py"
    conf["_py"] = sys.executable
    return conf


def mkdirs(path, conf):
    """
    mkdirs for "inputs_dir", "outputs_dir", "check_inputs_dir", "check_outputs_dir",
    if they are mentioned in conf dict

    :param path: str. full path for subway project.
    :param conf: Dict. config dict for subway project.
    :return: None
    """
    os.mkdir(os.path.join(path, ".subway"))
    for d in ["inputs_dir", "outputs_dir", "check_inputs_dir", "check_outputs_dir"]:
        _mkdir(path, conf, d)


def _mkdir(path, conf, key):
    if conf.get(key):
        os.mkdir(os.path.join(path, conf.get(key)))


def render_config(path, conf):
    """
    generate config.json based on conf dict within .subway dir

    :param path: str. full path for subway project.
    :param conf: Dict. config dict for subway project.
    :return: None
    """
    with open(os.path.join(path, ".subway", "config.json"), "w") as f:
        json.dump(conf, f, indent=2)


def render_history(path):
    """
    generate empty history.json with only ``{}`` within .subway dir

    :param path: str. full path for subway project.
    :return: None
    """
    with open(os.path.join(path, ".subway", "history.json"), "w") as f:
        json.dump({}, f, indent=2)


def render_main(path, conf, fromfile=None):
    """
    render main.py as detailed as possible and make its permission 700 (executable).

    :param path: str. full path for subway project.
    :param conf: Dict. config dict for subway project.
    :param fromfile: Optional[str]. Default None, indicates a default example main.py will be rendered.
                If specified as a full path, the corresponding file will be copied into subway project
                as entry point.
    :return: None.
    """
    if not conf:
        entry_point = "main.py"
    else:
        entry_point = conf.get("entry_point", "main.py")
    if not fromfile:
        mainpy = """#! {_py}

import os

from subway import cons

cons.work_dir = os.path.dirname(os.path.abspath(__file__))

from subway.config import conf, history

from subway.workflow import main_once, main_rt

from subway.plugins import DebugSub, DebugChk



if __name__ == "__main__":
    main_once(DebugChk(), DebugSub())

    """.format(
            _py=sys.executable
        )

        with open(os.path.join(path, entry_point), "w") as f:
            f.writelines([mainpy])
    else:  # fromfile
        shutil.copyfile(fromfile, os.path.join(path, entry_point))
    os.chmod(os.path.join(path, entry_point), 0o700)



"""
import config from ``config.json``
"""

import os
import json
from .cons import work_dir
from .utils import load_json


config_path = os.path.join(work_dir, ".subway", "config.json")
conf = load_json(config_path)

assert conf.get("work_dir", work_dir) == work_dir

conf["work_dir"] = work_dir
conf["config_path"] = config_path
conf["history_path"] = os.path.join(work_dir, ".subway", "history.json")
history = load_json(conf["history_path"])


def _conf_abs_dir(prefix):
    relkey = prefix + "_dir"
    abskey = prefix + "_abs_dir"
    if conf.get(relkey):
        conf[abskey] = os.path.join(conf["work_dir"], conf[relkey])


for prefix in ["inputs", "outputs", "check_inputs", "check_outputs"]:
    _conf_abs_dir(prefix)


def update_history():
    """
    write ``history`` in memory back to ``history.json``.

    :return: None.
    """
    with open(conf["history_path"], "w") as f:
        json.dump(history, f, indent=2)


def update_conf():
    """
    write ``conf`` in memory back to ``config.json``.

    :return: None.
    """
    with open(conf["config_path"], "w") as f:
        json.dump(conf, f, indent=2)



"""
Exception classes for subway.
"""
# TODO: more organized and hierachical exceptions.


class SubwayException(Exception):
    def __init__(self, message, code=10):
        """

        :param message: str.
        :param code: int. 10 general, 11 jobid unmatch, 12 only valid for general without id
            13 no such atrribute in history of conf
        """
        self.message = message
        self.code = code

    def __str__(self):
        return "%s %s" % (self.__class__.__name__, self.message)


class CLIException(SubwayException):
    def __init__(self, message, code=10):
        super().__init__(message, code)


class MatchError(CLIException):
    def __init__(self, message, code=11):
        super().__init__(message, code)


class NoAttribute(CLIException):
    def __init__(self, message, code=13):
        super().__init__(message, code)


class EndingBubble(SubwayException):
    pass



import os
from uuid import uuid4
from shutil import copyfile

from ..framework import PlainChk, PlainSub
from ..config import conf


class RgSub(PlainSub):
    def __init__(self, resource_limit=None, _py="python3"):
        self._py = _py
        super().__init__(resource_limit=resource_limit)

    def submit_pending(self, jobid):
        mainp = os.path.join(conf["work_dir"], conf["executable"])
        # shall use ./executables/randomg_run.py
        inputp = os.path.join(conf["inputs_abs_dir"], jobid)
        outputp = os.path.join(conf["outputs_abs_dir"], jobid)
        command = "%s %s %s %s" % (self._py, mainp, inputp, outputp)
        print("run %s" % command)
        os.system(command)


class RgChk(PlainChk):
    def check_kickstart(self):
        # newinput = str(uuid1())
        # L, l = self.params
        # print(L, l)
        # with open(os.path.join(conf["inputs_abs_dir"], newinput), "w") as f:
        #     f.writelines(["%s\n%s" % (L, l)])
        #
        # print("kickstart input")
        # return [(newinput, {})]
        r = []
        for param in self.params:
            nid = str(uuid4())
            L, l = param
            with open(os.path.join(conf["inputs_abs_dir"], nid), "w") as f:
                f.writelines(["%s\n%s" % (L, l)])
            r.append((nid, {}))
        return r

    def check_checking(self, outputpath):
        with open(os.path.join(conf["inputs_abs_dir"], outputpath), "r") as f:
            L, l = f.readlines()
        l = float(l)
        with open(os.path.join(conf["outputs_abs_dir"], outputpath), "r") as f:
            r = f.readlines()
        r = float(r[0])
        if r < l:
            print("find the converged result, computation stopped for this parameter")
            return []

        newinput = str(uuid4())
        print("begin copy file...")
        copyfile(
            os.path.join(conf["inputs_abs_dir"], outputpath),
            os.path.join(conf["inputs_abs_dir"], newinput),
        )
        return [(newinput, {})]



from setuptools import setup, find_packages
from subway import version, __description__, __license__


with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()


if __name__ == "__main__":
    setup(
        name="hpcsubway",
        version=version,
        packages=find_packages(),
        description=__description__,
        license=__license__,
        long_description=readme,
        long_description_content_type="text/markdown",
        python_requires=">=3.5",
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
        ],
        entry_points={"console_scripts": ["subway = subway.entry:cli"]},
    )



import os

from ..plugins import SSlurmChk, SSlurmSub
from ..config import conf


class RgSSub(SSlurmSub):
    pass


class RgSChk(SSlurmChk):
    def check_checking_main(self, jobid):
        with open(os.path.join(conf["inputs_abs_dir"], jobid), "r") as f:
            L, l = f.readlines()
        L = float(L)
        l = float(l)
        with open(os.path.join(conf["outputs_abs_dir"], jobid), "r") as f:
            r = f.readlines()
        r = float(r[0])
        if r < l:
            print(
                "find the converged result, computation stopped for this parameter %s"
                % jobid
            )
            return []  # no new calculation is needed
        return [[L, l]]



"""
Main loop function for subway entry_point
"""

import time
import sys

from .exceptions import EndingBubble


def main_once(checker, submitter):
    """
    main entry point run checker and submitter once.

    :param checker: Checker.
    :param submitter: Submitter.
    :return: None
    """

    ######
    # check part
    ######

    print("------begin checker------")
    checker()

    ######
    # submit part
    ######

    print("------begin submitter------")
    submitter()


def main_rt(
    checker,
    submitter,
    preprocessor=None,
    postprocessor=None,
    sleep_interval=10,
    loops=None,
):
    """
    realtime main monitor for subway entry_point

    :param checker: Checker.
    :param submitter: Submitter.
    :param preprocessor: Optional[Processor]. Run before the main loop.
    :param postprocessor: Optional[Processor]. Run after the main loop.
    :param sleep_interval: Optional[Union[float, Tuple[float, float]]], default 10.
                There are two intervals, one after checker, one after submitter.
                If one number is given, then two intervals are the same.
                If tuple or list of 2 numbers are given, they are explained as
                (after_checker_interval, after_submitter_interval)
    :param loops: Optional[int], default None.
                max loops before exit, none for running forever,
                but the monitor can still quit if no job is running as expected.
    :return: None
    """
    # TODO: better way for realtime always-on monitor version of main?
    i = 0
    if isinstance(sleep_interval, tuple) or isinstance(sleep_interval, list):
        after_checker_interval, after_submitter_interval = sleep_interval
    else:
        after_checker_interval = after_submitter_interval = sleep_interval
    if preprocessor:
        preprocessor()
    try:
        while (not loops) or (i < loops):
            print("------begin checker------")
            checker()
            time.sleep(after_checker_interval)
            print("------begin submitter------")
            submitter()
            time.sleep(after_submitter_interval)
            i += 1
    except KeyboardInterrupt as e:
        print("Quit the main monitor", file=sys.stderr)
    except EndingBubble:
        print("No active jobs anymore, quitting", file=sys.stderr)
    finally:
        if postprocessor:
            postprocessor()



from ..framework import PlainChk, PlainSub
from subway.components import SlurmJob
from ..config import history
from ..utils import now_ts
from ..components.slurmoo import slurm_abnormal_states


class SlurmChk(PlainChk):
    def is_finished(self, jobid):
        """
        This utils is to further distinguish those tasks that has incomplete outfile while still running

        :param jobid: relativepath
        :return:
        """
        sjob = SlurmJob(jobname=jobid)
        ifn = sjob.jobinfo["State"] == "COMPLETED"
        if ifn:
            history[jobid]["beginning_real_ts"] = sjob.jobinfo["Start_ts"]
            return True
        return False

    def is_aborted(self, jobid):
        sjob = SlurmJob(jobname=jobid)
        if sjob.jobinfo["State"] in slurm_abnormal_states:
            history[jobid]["reason"] = sjob.jobinfo["State"]
            return True
        return False

    def is_checked(self, jobid):
        jid = getattr(history[jobid], "assoc", "")  # jobname for associate check job
        if jid:
            sjob = SlurmJob(jobname=jid)
            return sjob.jobinfo["State"] == "COMPLETED"
        # no independent checking job
        return True

    def is_frustrated(self, jobid):
        jid = getattr(history[jobid], "assoc", "")  # jobname for associate check job
        if jid:
            sjob = SlurmJob(jobname=jid)
            if sjob.jobinfo["State"] in slurm_abnormal_states:
                history[jobid]["check_reason"] = sjob.jobinfo["State"]
                return True
            return False
        # no independent checking job, always checked instead of frustrated
        return False

    def is_resolved(self, jobid):
        self.is_checked(jobid)

    def is_failed(self, jobid):
        self.is_frustrated(jobid)

    def finishing_time(self, jobid):
        """


        :param jobid:
        :return:
        """
        sjob = SlurmJob(jobname=jobid)

        return sjob.jobinfo["End_ts"]

    def ending_time(self, jobid):
        """
        ending_time is reponsible for 4 states: checked resolved frustrated failed

        :param jobid:
        :return:
        """
        jid = getattr(history[jobid], "assoc", "")  # jobname for associate check job
        if jid:
            sjob = SlurmJob(jobname=jid)
            return sjob.jobinfo["End_ts"]
        return now_ts()


class SlurmSub(PlainSub):
    pass
